/*
libutil -- read config file

Written by Arnt Gulbrandsen <agulbra@troll.no> and copyright 1995
Troll Tech AS, Postboks 6133 Etterstad, 0602 Oslo, Norway, fax +47
22646949.
Modified by Cornelius Krasel <krasel@wpxx02.toxi.uni-wuerzburg.de>
and Randolf Skerka <Randolf.Skerka@gmx.de>.
Copyright of the modifications 1997.
Modified by Kent Robotti <robotti@erols.com>. Copyright of the
modifications 1998.
Modified by Markus Enzenberger <enz@cip.physik.uni-muenchen.de>.
Copyright of the modifications 1998.
Modified by Cornelius Krasel <krasel@wpxx02.toxi.uni-wuerzburg.de>.
Copyright of the modifications 1998, 1999.
Modified and copyright of the modifications 2001 - 2002 by Matthias Andree
<matthias.andree@web.de>.
Modified and copyright of the modifications 2002 by Ralf Wildenhues
<ralf.wildenhues@gmx.de>.

See file COPYING for restrictions on the use of this software.
*/

#include "leafnode.h"
#include "validatefqdn.h"
#include "ln_log.h"
#include "strlcpy.h"

#include <ctype.h>
#include <errno.h>
#include <limits.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <syslog.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <time.h>
#include <sys/resource.h>

#define COREFILESIZE 1024*1024*16
#define TOKENSIZE 1024

#include "groupselect.h"

/*
 * misc. global variables, documented in leafnode.h
 */
time_t expire = 0;
int expiredays = 0;
struct expire_entry *expire_base;
unsigned long artlimit = 0;
unsigned long initiallimit = 0;
long crosspostlimit = 0;
int create_all_links = 0;
int delaybody = 0;
int debugmode = 0;		/* if 1, log lots of stuff via syslog */
int maxage = 10;
long maxlines = 0;
long minlines = 0;
unsigned long maxbytes = 0;
int linebuffer = 0;		/* if 1, make stdout and stderr explicitly 
				   line buffered, GNU libc makes them fully buffered
				   if redirected to files */
int timeout_long = 7;
int timeout_short = 2;
int timeout_active = 90;
int timeout_client = 15*60;
int clamp_maxage = 1;           /* if 1, maxage will be lowered to
				 * groupexpire or expire if the
				 * applicable parameter is lower than
				 * maxage, to prevent duplicate fetches
				 * after a premature exit of
				 * fetchnews. */
int allowstrangers = 0;
char *filterfile;
struct server *servers = NULL;
int allow_8bit_headers = 0;

static int parse_line( /*@unique@*/ char *l, /*@out@*/ char *param,	/*@out@*/
		      char *value);

/* parse a line, destructively */
int
parse_line(char *l, char *param, char *value)
{
    char *p, *q;

    p = l;
    /* skip comments */
    q = strchr(p, '#');
    if (q)
	*q = '\0';
    if (*p == '\0')
	return 0;
    if ((q = strchr(p, '=')) == NULL)
	return 0;
    else
	*q++ = '\0';
    /* skipping leading blanks or tabs */
    SKIPLWS(p);
    SKIPLWS(q);
    (void)strlcpy(param, p, TOKENSIZE);
    (void)strlcpy(value, q, TOKENSIZE);

    /* now param contains the stuff before '=' and value the stuff behind it */
    if (!(*param) || !(*value))
	return 0;
    /* skipping trailing blanks or tabs */
    p = param + strlen(param) - 1;
    q = value + strlen(value) - 1;
    while (isspace((unsigned char)*p) && (p > param)) {
	*p = '\0';
	p--;
    }
    while (isspace((unsigned char)*q) && (q > value)) {
	*q = '\0';
	q--;
    }
    return 1;
}

/*
05/25/97 - T. Sweeney - Modified to read user name and password for AUTHINFO.
                        Security questionable as password is stored in
                        plaintext in insecure file.
1999-07-15 - Matthias Andree
             Set p and q defaults to 0
*/

int
readconfig(int logtostderr)
{
    struct server *p = 0, *q = 0;
    struct rlimit corelimit;
    struct expire_entry *ent = NULL, *prev = NULL;
    FILE *f;
    char *l;
    char *param, *value;
    char s[SIZE_s + 1];
    unsigned long curline = 0;

    artlimit = 0;
    param = critmalloc(TOKENSIZE, "allocating space for parsing");
    value = critmalloc(TOKENSIZE, "allocating space for parsing");

    xsnprintf(s, SIZE_s, "%s/config", sysconfdir);
    if ((f = fopen(s, "r")) == NULL) {
	syslog(LOG_ERR, "cannot open %s", s);
	free(param);
	free(value);
	return 0;
    }
    while ((l = getaline(f))) {
	++curline;
	if (parse_line(l, param, value)) {
	    if (strcmp("username", param) == 0) {
		if (p) {
		    if (p->username != NULL)
			free(p->username);
		    p->username = strdup(value);
		    if (debugmode)
			syslog(LOG_DEBUG, "config: found username for %s",
			       p->name);
		} else
		    syslog(LOG_ERR, "config: no server for username %s", value);
	    } else if (strcmp("password", param) == 0) {
		if (p) {
		    if (p->password != NULL)
			free(p->password);
		    p->password = strdup(value);
		    if (debugmode)
			syslog(LOG_DEBUG, "config: found password for %s",
			       p->name);
		} else
		    syslog(LOG_ERR, "config: no server for password");
	    } else if (strcmp("timeout", param) == 0) {
		if (p) {
		    p->timeout = atoi(value);
		    if (debugmode)
			syslog(LOG_DEBUG, "config: timeout is %d seconds",
			       p->timeout);
		} else
		    syslog(LOG_ERR, "config: no server for timeout");
	    } else if (strcmp("allowSTRANGERS", param) == 0) {
		if (value && strlen(value)) {
		    if (atoi(value) == 42)
			allowstrangers = 1;
		    syslog(LOG_DEBUG,
			    "config: allowstrangers is %s",
			    allowstrangers ? "set" : "unset");
		}
	    } else if (strcmp("create_all_links", param) == 0) {
		if (value && strlen(value)) {
		    create_all_links = atoi(value);
		    if (create_all_links && debugmode)
			syslog(LOG_DEBUG,
				"config: link articles in all groups");
		}
	    } else if (strcmp("expire", param) == 0) {
		int i = atoi(value);
		if (i >= (INT_MAX / SECONDS_PER_DAY))
		    i = (INT_MAX / SECONDS_PER_DAY) - 1;
		if (i <= 0) {
		    ln_log(LNLOG_SERR, LNLOG_CTOP, "config: expire must be positive, not %d, abort", i);
		    exit(1);
		}
		expiredays = i;
		expire = time(NULL) - (time_t) (SECONDS_PER_DAY * i);
		if (debugmode)
		    syslog(LOG_DEBUG, "config: expire is %d days", i);
	    } else if (strcmp("filterfile", param) == 0) {
		if (debugmode)
		    syslog(LOG_DEBUG, "config: filterfile is %s", value);
		filterfile = strdup(value);
	    } else if ((strcmp("hostname", param) == 0) ||
		    (strcmp("fqdn", param) == 0)) {
		if (is_validfqdn(value)) {
		    if (debugmode)
			syslog(LOG_DEBUG, "config: hostname is %s", value);
		    xstrlcpy(fqdn, value, sizeof(fqdn));
		} else {
		    ln_log(LNLOG_SCRIT, LNLOG_CTOP, "config: hostname %s is "
			    "not a valid fully-qualified domain name, exit.",
			    value);
		    exit(1);
		}
	    } else if ((strcmp("maxcrosspost", param) == 0) ||
		       (strcmp("maxgroups", param) == 0)) {
		/* maxgroups is for compatibility with leafnode+ */
		crosspostlimit = strtol(value, NULL, 10);
		if (debugmode)
		    syslog(LOG_DEBUG, "config: crosspostlimit is %ld groups",
			   crosspostlimit);
	    } else if (strcmp("clamp_maxage", param) == 0) {
		clamp_maxage = atoi(value);
		if (debugmode)
		    syslog(LOG_DEBUG, "config: clamp_maxage is %s",
			   clamp_maxage ? "TRUE" : "FALSE");
	    } else if (strcmp("maxlines", param) == 0) {
		maxlines = strtol(value, NULL, 10);
		if (debugmode)
		    syslog(LOG_DEBUG, "config: postings have max. %ld lines",
			   maxlines);
	    } else if (strcmp("minlines", param) == 0) {
		minlines = strtol(value, NULL, 10);
		if (debugmode)
		    syslog(LOG_DEBUG, "config: postings have min. %ld lines",
			   minlines);
	    } else if (strcmp("maxbytes", param) == 0) {
		maxbytes = strtol(value, NULL, 10);
		if (debugmode)
		    syslog(LOG_DEBUG,
			   "config: postings have max. %lu bytes", maxbytes);
	    } else if (strcmp("linebuffer", param) == 0) {
		linebuffer = atoi(value);
		if (debugmode)
		    syslog(LOG_DEBUG, "config: linebuffer is %d", linebuffer);
	    } else if (strcmp("allow_8bit_headers", param) == 0) {
		allow_8bit_headers = atoi(value);
		if (debugmode)
		    syslog(LOG_DEBUG, "config: allow_8bit_headers is %d",
			    allow_8bit_headers);
	    } else if (strcmp("debugmode", param) == 0) {
		debugmode = atoi(value);
		if (debugmode)
		    syslog(LOG_DEBUG, "config: debugmode is %d", debugmode);
	    } else if (strcmp("delaybody", param) == 0) {
		delaybody = atoi(value);
		if (debugmode)
		    syslog(LOG_DEBUG, "config: delaybody is %d (default 0)",
			   delaybody);
	    } else if (strcmp("timeout_short", param) == 0) {
		timeout_short = atoi(value);
		if (debugmode)
		    syslog(LOG_DEBUG, "config: timeout_short is %d days",
			   timeout_short);
	    } else if (strcmp("timeout_long", param) == 0) {
		timeout_long = atoi(value);
		if (debugmode)
		    syslog(LOG_DEBUG, "config: timeout_long is %d days",
			   timeout_long);
	    } else if (strcmp("timeout_active", param) == 0) {
		timeout_active = atoi(value);
		if (debugmode)
		    syslog(LOG_DEBUG, "config: timeout_active is %d days",
			   timeout_active);
	    } else if (strcmp("timeout_client", param) == 0) {
		timeout_client = atoi(value);
		if (debugmode)
		    syslog(LOG_DEBUG, "config: timeout_client is %d secs",
			   timeout_client);
	    } else if (strncmp("groupexpire", param, 11) == 0) {
		char *m;
		m = param;
		while (*m && !(isspace((unsigned char)*m)))
		    m++;
		while (isspace((unsigned char)*m))
		    m++;
		if (m && *m) {
		    time_t e, i = (time_t) atol(value);
		    if (i >= (INT_MAX / SECONDS_PER_DAY))
			i = (INT_MAX / SECONDS_PER_DAY) - 1;
		    if (debugmode) {
			if ((long)i < 0)
			    syslog(LOG_DEBUG,
			           "config: groupexpire for %s is %ld (never)",
				   m, (long)i);
			else
			    syslog(LOG_DEBUG,
			           "config: groupexpire for %s is %ld days",
			           m, (long)i);
		    }
		    e = time(NULL) - (time_t) (SECONDS_PER_DAY * i);
		    ent = (struct expire_entry *)
			critmalloc(sizeof(struct expire_entry), "readconfig");
		    ent->group = critstrdup(m, "readconfig");
		    ent->days = i;
		    ent->xtime = e;
		    ent->next = prev;
		    prev = ent;
		}
	    } else if ((strcmp("maxage", param) == 0) ||
		       (strcmp("maxold", param) == 0)) {
		/* maxold is for compatibility with leafnode+ */
		maxage = atoi(value);
		if (debugmode)
		    syslog(LOG_DEBUG, "config: maxage is %d", maxage);
	    } else if (strcmp("maxfetch", param) == 0) {
		artlimit = strtoul(value, NULL, 10);
		if (debugmode)
		    syslog(LOG_DEBUG, "config: maxfetch is %lu", artlimit);
	    } else if (strcmp("port", param) == 0) {
		unsigned long pp = strtoul(value, NULL, 10);
		if (p) {
		    if (pp == 0 || pp > 65535) {
			syslog(LOG_ERR,
			       "config: invalid port number for nntpport %s",
			       value);
		    } else {
			p->port = (unsigned int)pp;
			if (debugmode)
			    syslog(LOG_DEBUG, "config: nntpport is %u",
				   p->port);
		    }
		} else {
		    syslog(LOG_ERR, "config: no server for nntpport %s", value);
		}
	    } else if (strcmp("noactive", param) == 0) {
		if (p) {
		    p->updateactive = FALSE;
		    if (debugmode)
			syslog(LOG_DEBUG, "config: no active file updates for %s",
			       p->name);
		} else 
		    syslog(LOG_ERR, "config: no server for noactive = %s", 
			   value);
	    } else if (strcmp("nodesc", param) == 0) {
		if (p) {
		    p->descriptions = FALSE;
		    if (debugmode)
			syslog(LOG_DEBUG, "config: no LIST NEWSGROUPS for %s",
			       p->name);
		} else
		    syslog(LOG_ERR, "config: no server for nodesc = %s", value);
	    } else if (strcmp("nopost", param) == 0) {
		if (p) {
		    p->nopost = atoi(value);
		    if (debugmode)
			syslog(LOG_DEBUG, "config: nopost for %s is %d",
			       p->name, p->nopost);
		} else
		    syslog(LOG_ERR, "config: no server for nopost = %s", value);
	    } else if (strcmp("initialfetch", param) == 0) {
		initiallimit = strtoul(value, NULL, 10);
		if (debugmode)
		    syslog(LOG_DEBUG, "config: initialfetch is %lu",
			   initiallimit);
	    } else if ((strcmp("server", param) == 0) ||
		       (strcmp("supplement", param) == 0)) {
		if (debugmode)
		    syslog(LOG_DEBUG, "config: server is %s", value);
		p = (struct server *)critmalloc(sizeof(struct server),
						    "allocating space for server name");
		p->name = critstrdup(value, "readconfig");
		p->descriptions = TRUE;
		p->next = NULL;
		p->timeout = 30;	/* default 30 seconds */
		p->port = 0;
		p->username = NULL;
		p->password = NULL;
		p->nopost = 0;
		p->updateactive = TRUE;
		p->group_pcre = NULL;
		if (servers == NULL)
		    servers = p;
		else
		    q->next = p;
		q = p;
	    } else if (0 == strcmp("only_groups_pcre", param)) {
		pcre *re = gs_compile(value);
		if (!re) exit(2);
		if (p) {
		    p->group_pcre = re;
		    if (debugmode)
			syslog(LOG_DEBUG, "config: only_groups_pcre for %s is %s",
			       p->name, value);
		} else {
		    syslog(LOG_ERR, "config: no server for nopost = %s", value);
		}
	    } else {
		ln_log(LNLOG_SERR, LNLOG_CTOP,
			"config: unknown line %lu: \"%s = %s\"", curline,
			param, value);
	    }
	} else {
	    size_t i;
	    if ((i = strspn(l, " \t")) < strlen(l) && l[i] != '#') {
		ln_log(LNLOG_SERR, LNLOG_CTOP,
			"config: malformatted line %lu: \"%s\"", curline,
			l);
	    }
	}
    }
    if (maxage != 0 && maxage > expiredays && clamp_maxage == 0) {
	ln_log(LNLOG_SERR, LNLOG_CTOP,
		"config: maxage (%d) > expire (%d). This can cause duplicate download. Please fix your configuration, maxage must not be greater than expire.",
		maxage, expiredays);
	exit(1);
    }
    debug = debugmode;

    expire_base = ent;
    fclose(f);
    if (servers == NULL) {
	syslog(LOG_ERR, "no server declaration in config file");
	free(param);
	free(value);
	return 0;
    }
    if (!expire)
	syslog(LOG_ERR, "no expire declaration in config file");

    if (debugmode > 1) {
	corelimit.rlim_cur = COREFILESIZE;
	if (setrlimit(RLIMIT_CORE, &corelimit) < 0)
	    syslog(LOG_DEBUG, "Changing core file size failed: %m");
	corelimit.rlim_cur = 0;
	getrlimit(RLIMIT_CORE, &corelimit);
	syslog(LOG_DEBUG, "Core file size: %d", (int)corelimit.rlim_cur);
    }

    if (linebuffer) {
	fflush(stdout);
	setvbuf(stdout, NULL, _IOLBF, BUFSIZ);
	fflush(stderr);
	setvbuf(stderr, NULL, _IOLBF, BUFSIZ);
    }

    free(param);
    free(value);
    validatefqdn(logtostderr);
    return 1;
}

/*
1997-05-27 - T. Sweeney - Find a group in the expireinfo linked list and return
                          its expire time. Otherwise, return zero.
*/
static struct expire_entry *
lookup_expireent(char *group)
{
    struct expire_entry *a;

    a = expire_base;
    while (a != NULL) {
	if (ngmatch(a->group, group) == 0)
	    return a;
	a = a->next;
    }
    return NULL;
}

static void
freeserver(struct server *a) {
    if (a->name) free(a->name);
    if (a->username) free(a->username);
    if (a->password) free(a->password);
}

void
freeservers(void) {
    struct server *i = servers, *n;

    while(i != NULL) {
	n = i->next;
	freeserver(i);
	i = n;
    }
}

time_t lookup_expire(char *group)
{
    struct expire_entry *e;
    e = lookup_expireent(group);
    if (e) return e->xtime;
    return 0;
}

int lookup_expiredays(char *group)
{
    struct expire_entry *e;
    e = lookup_expireent(group);
    if (e) return e->days;
    return 0;
}
